import click
import re
from click import Abort
from imagination import container
from typing import Optional, Any, Dict
from urllib.parse import urlparse

from dnastack.cli.data_connect.helper import handle_query
from dnastack.cli.helpers.client_factory import ConfigurationBasedClientFactory
from dnastack.cli.helpers.command.decorator import command
from dnastack.cli.helpers.command.spec import ArgumentSpec, RESOURCE_OUTPUT_SPEC, DATA_OUTPUT_SPEC
from dnastack.cli.helpers.command.group import AliasedGroup
from dnastack.cli.helpers.exporter import normalize, to_json
from dnastack.cli.helpers.iterator_printer import show_iterator
from dnastack.client.collections.client import CollectionServiceClient
from dnastack.client.data_connect import DataConnectClient
from dnastack.common.logger import get_logger


def _get(id: Optional[str] = None) -> CollectionServiceClient:
    factory: ConfigurationBasedClientFactory = container.get(ConfigurationBasedClientFactory)
    return factory.get(CollectionServiceClient, id)


def _switch_to_data_connect(collection_service_client: CollectionServiceClient,
                            collection_id_or_slug_name: Optional[str],
                            no_auth: bool) -> DataConnectClient:
    try:
        return DataConnectClient.make(collection_service_client.data_connect_endpoint(collection_id_or_slug_name))
    except AssertionError:
        collection_service_client.list_collections(no_auth)
        _abort_with_collection_list(collection_service_client, collection_id_or_slug_name, no_auth=no_auth)


def _abort_with_collection_list(collection_service_client: CollectionServiceClient,
                                collection_id_or_slug_name: Optional[str],
                                no_auth: bool):
    available_identifiers = "\n - ".join(sorted([
        available_collection.slugName
        for available_collection in collection_service_client.list_collections(no_auth=no_auth)
    ]))

    error_message = f'The collection ID or slug name is not given via --collection or -c option or it is invalid. ' \
                    f'(Given: {collection_id_or_slug_name})'

    if available_identifiers:
        raise Abort(f'{error_message}\n\nHere is the list of available collection IDs:\n\n'
                    f' - {available_identifiers}\n')
    else:
        raise Abort(f'{error_message}\n\nHowever, you do not seem to have access to any collection at '
                    f'{collection_service_client.url} or there are no collections available at the moment.\n')


@click.group("collections", cls=AliasedGroup, aliases=['cs'])
def collection_command_group():
    """ Interact with Collection Service or Explorer Service (e.g., Viral AI) """


@command(collection_command_group, 'list', specs=[RESOURCE_OUTPUT_SPEC])
def list_collections(endpoint_id: Optional[str], no_auth: bool = False, output: Optional[str] = None):
    """ List collections """
    show_iterator(output, _get(endpoint_id).list_collections(no_auth=no_auth))


@command(collection_command_group,
         specs=[
             ArgumentSpec(
                 name='collection',
                 arg_names=['--collection', '-c'],
                 as_option=True,
                 help='The ID or slug name of the target collection; required only by an explorer service',
             ),
             ArgumentSpec(
                 name='limit',
                 arg_names=['--limit', '-l'],
                 as_option=True,
                 help='The maximum number of items to display',
             ),
             RESOURCE_OUTPUT_SPEC
         ])
def list_items(endpoint_id: Optional[str],
               collection: Optional[str],
               limit: Optional[int] = 50,
               no_auth: bool = False,
               output: Optional[str] = None):
    """ List items of the given collection """
    logger = get_logger('CLI/list-items')
    limit_override = False

    limit = int(limit)  # This is for Python 3.7.

    collection_service_client = _get(endpoint_id)

    collection_id = collection.strip() if collection else None
    if not collection_id:
        _abort_with_collection_list(collection_service_client, collection, no_auth=no_auth)

    actual_collection = collection_service_client.get(collection_id, no_auth=no_auth)
    data_connect_client = _switch_to_data_connect(collection_service_client, actual_collection.slugName, no_auth)
    items_query = actual_collection.itemsQuery.strip()

    if re.search(r' limit\s*\d+$', items_query, re.IGNORECASE):
        logger.warning('The items query already has the limit defined and the CLI will not override that limit.')
    else:
        logger.debug(f'Only shows {limit} row(s)')
        limit_override = True
        items_query = f'{items_query} LIMIT {limit + 1}'  # We use +1 as an indicator whether there are more results.

    def __simplify_item(row: Dict[str, Any]) -> Dict[str, Any]:
        # NOTE: It is implemented this way to guarantee that "id" and "name" are more likely to show first.
        property_names = ['type', 'size', 'size_unit', 'version', 'item_updated_at']

        item = dict(
            id=row['id'],
            name=row.get('qualified_table_name') or row['preferred_name'],
        )

        if row['type'] == 'blob':
            property_names.extend([
                'checksums',
                'metadata_url',
                'mime_type',
            ])
        elif row['type'] == 'table':
            property_names.extend([
                'json_schema',
            ])

        item.update({
            k: v
            for k, v in row.items()
            if k in property_names
        })

        # FIXME: Remove this logic when https://www.pivotaltracker.com/story/show/182309558 is resolved.
        if 'metadata_url' in item:
            parsed_url = urlparse(item['metadata_url'])
            item['metadata_url'] = f'{parsed_url.scheme}://{parsed_url.netloc}/{item["id"]}'

        return item

    items = [i for i in data_connect_client.query(items_query, no_auth=no_auth)]
    row_count = len(items)

    show_iterator(
        output or RESOURCE_OUTPUT_SPEC.default,
        items,
        __simplify_item,
        limit
    )

    click.secho(f'Displayed {row_count} item{"s" if row_count != 1 else ""} from this collection',
                fg='green',
                err=True)

    if limit_override and row_count > limit:
        click.secho(f'There exists more than {limit} item{"s" if row_count != 1 else ""} in this collection.',
                    dim=True,
                    err=True)


@command(collection_command_group,
         'query',
         [
             ArgumentSpec(
                 name='collection',
                 arg_names=['--collection', '-c'],
                 as_option=True,
                 help='The ID or slug name of the target collection; required only by an explorer service',
             ),
             ArgumentSpec(
                 name='decimal_as',
                 arg_names=['--decimal-as'],
                 as_option=True,
                 help='The format of the decimal value',
                 choices=["string", "float"],
             ),
             DATA_OUTPUT_SPEC
         ])
def query_collection(endpoint_id: Optional[str],
                     collection: Optional[str],
                     query: str,
                     decimal_as: str = 'string',
                     no_auth: bool = False,
                     output: Optional[str] = None):
    """ Query data """
    client = _switch_to_data_connect(_get(endpoint_id), collection, no_auth=no_auth)
    return handle_query(client, query, decimal_as=decimal_as, no_auth=no_auth, output_format=output)


@click.group("tables")
def table_command_group():
    """ Data Client API for Collections """


@command(table_command_group,
         'list',
         [
             ArgumentSpec(
                 name='collection',
                 arg_names=['--collection', '-c'],
                 as_option=True,
                 help='The ID or slug name of the target collection; required only by an explorer service',
             ),
         ])
def list_tables(endpoint_id: Optional[str], collection: Optional[str], no_auth: bool = False):
    """ List all accessible tables """
    client = _switch_to_data_connect(_get(endpoint_id), collection, no_auth=no_auth)
    click.echo(to_json([t.dict() for t in client.list_tables(no_auth=no_auth)]))


# noinspection PyTypeChecker
collection_command_group.add_command(table_command_group)
